//----------------------------------------------------------------------------------------------------------------------
// Tio	MANAGE TEXT I/O  Created on 24 Feb 2012		Copyright Geoff Kaniuk 2012
//----------------------------------------------------------------------------------------------------------------------
//  This file is part of toursim.
//
//  toursim is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  toursim is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with toursim.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------------------------------------------------
// A copy of the GNU General Public License is released with this software in the file COPYING
//----------------------------------------------------------------------------------------------------------------------
#include <cctype>
#include <sys/stat.h>

#include "tio.h"

//----------------------------------------------------------------------------------------------------------------------
// ERROR REPORTING
//
namespace Tio
{
	using std::cout; using std::endl;	using std::clog;

	enum
	{
		ep_makespecs = ep_nerrtioh,											// Continue from tio.h
		ep_makeblock, ep_makeblocks,
		ep_iname, ep_inoname, ep_badclass, ep_badcount, ep_badspec,

		ep_bname, ep_bzname,
		ep_anonymousblock, ep_namedblocks, ep_emptyblocks,
		ep_emptyblockname, ep_missingblock,

		ep_itemindex, ep_blockindex, ep_blocksize, ep_erase,

		ep_items, ep_finditem,
		ep_chooser, ep_noklist, ep_idkeylist, ep_itemkey,
		ep_kimport, ep_badmask,
		ep_badchoice, ep_badchoices,


		ep_garbage, ep_class, ep_fieldlist, ep_item,

		ep_getline, ep_syntax,
		ep_scalar, ep_interval, ep_pair, ep_row, ep_vector, ep_matrix,

		ep_loaditem,

		ep_unknown
	};

	const char* tiomsg[ep_unknown+1] =
	{																		// Errors in tio.h or extern
	 	 "no error",				// ep_noerror
	 	 "vector dim",				// ep_vectordim
	 	 "matrix dim",				// ep_matrixdim
	 	 "empty filename",			// ep_fname
	 	 "no such file",			// ep_fexist
	 	 "cannot open file",		// ep_fopen
	 	 "invalid file header",     // ep_fheader
	 	 "premature file end",		// ep_fend
	 	 "cannot close file",		// ep_fclose
	 	 	 	 	 	 	 	 	 	 	 	 	 	 	 	 	 	 	// Errors local to tio.cpp
	 	 "itemspecs construct",		// ep_makespecs
	 	 "itemsblock construct",	// ep_makeblock,
	 	 "itemblocks construct",	// ep_makeblocks

	 	 "repeated item name",		// ep_iname
	 	 "empty item name",			// ep_inoname
	 	 "unknown class", 			// ep_badclass
	 	 "invalid field count",		// ep_badcount
	 	 "invalid Itemspec",		// ep_badspec

	 	 "repeated block name",		// ep_bname
	 	 "empty block name",		// ep_bzname
	 	 "bad anonymous block",		// ep_anonymousblock
	 	 "bad named blocks",		// ep_namedblocks
	 	 "empty blocks",			// ep_emptyblocks
	 	 "empty blockname required",// ep_emptyblockname,
	 	 "cannot find block name",  // ep_missingblock

	 	 "item index out of range", // ep_itemindex
	 	 "block index out of range",// ep_blockindex
	 	 "illegal block size",		// ep_blocksize
	 	 "illegal items erase",		// ep_erase

	 	 "corrupt items string",	// ep_items
	 	 "cannot find item spec",	// ep_finditem
	 	 "expected choice/choices", // ep_chooser
	 	 "empty keylist",			// ep_linkclass
	 	 "bad keylist id",			// ep_idkeylist
	 	 "item key mismatch",		// ep_itemkey
	 	 "bad keylist import",		// ep_kimport
	 	 "bad mask",				// ep_badmask
	 	 "invalid choice value",	// ep_badchoice
	 	 "invalid choices value",	// ep_badchoices



	 	 "garbage follows data",	// ep_garbage
	 	 "unknown or wrong class",	// ep_class
	 	 "fields dont match spec",	// ep_fieldlist
	 	 "item never loaded",		// ep_item

	 	 "cannot read line",		// ep_getline
	 	 "syntax error",			// ep_syntax

	 	 "scalar error",			// ep_scalar
	 	 "interval error",			// ep_interval
	 	 "pair error",				// ep_pair

	 	 "array error",				// ep_row
	 	 "vector error",			// ep_vector
	 	 "matrix error",			// ep_matrix,

	 	 "bad class in load_item",	// ep_loaditem

	 	 "unknown error"
	};

	const char* tio_error_message(int err)
	{
		if(err<0 || err>=ep_unknown) err = ep_unknown;

		return tiomsg[err];
	}
}

//----------------------------------------------------------------------------------------------------------------------
// STRING TOOLS
//
namespace Tio
{
	using std::string;

	using Appmaths::real;

	const string tab("\t"),  hash("#"),  hyphen("-"), space(" ");
	const string point("."), colon(":"), hcolon(";"), uscore("_");
	const string equals("="), plus("+"), star("*"),   query("?");
	const string vbar("|"),   ampers("&");

	const char* dummy_cstring("");

	string dummy_string;

	int str_to_int(const string& source)
	{
		std::istringstream istream(source, std::ios::in);

		int i;	istream>>i;

		return i;
	}

	real str_to_real(const string& source)
	{
		std::istringstream rstream(source, std::ios::in);

		real x;	rstream>>x;

		return x;
	}

	void trim(string& source)
	{
		if(source.empty()) return;											// Left trim

		string::iterator f = source.begin(), e = source.end();

		while(f != e)
		{
			if(!isspace(*f)) break;	++f;
		}

		if(f == e) { source.clear(); return; }

		--e;																// Right trim

		while(e != f)
		{
			if(!isspace(*e)) break;	--e;
		}

		source = string(f,++e);												// At least one non-whitespace character
	}

	bool is_whitespace(const char c) { return c<32 or c==' '; }

	bool parse(const string& source, const char delim, vector<string>& sink)
	{
		const char* csource = source.c_str();

		if(not csource or is_whitespace(delim)) return false;

		const char *isrce = csource, *esrce = csource + source.size();
		const char *fsrce = csource;

		sink.clear();

		while(isrce != esrce)
		{
			if(*isrce==delim)
			{
				if(isrce==fsrce) return false;								// Cannot start with delim

				sink.push_back(string(fsrce, isrce));						// Extract item

				fsrce = isrce +1;
			}
			++isrce;
		}
		if(fsrce==esrce) return false;										// Cannot end with delim

		sink.push_back(string(fsrce, esrce));								// Extract last item

		return true;
	}
}

//----------------------------------------------------------------------------------------------------------------------
// KEY LISTS AND CHOICES
//
namespace Tio
{
	const Keylist dummy_keylist;
	const Keylists dummy_keylists;
	const Refkeylists dummy_refkeylists;

	enum { kl_none, kl_all, kl_nextra};
	const char* kl_extras[kl_nextra] = { "none", "all" };

	bool Choices::build_mask(const Keylist& keylist, const string& kselects)
	{
		nkey = keylist.nkey;	if(nkey<=0) return false;

		ready = false;

		vector<string>selects;

		if(not parse(kselects, '|', selects))return false;					// Extract selections

		bitmask = 0;	finger = 1;

		for(size_t k=0; k<selects.size(); k++)
		{
			const string& kselect(selects[k]);								// The next selection

			size_t kid = nkey;

			for(kid=0; kid<nkey; kid++)										// Find kselect in keys
				if(kselect==string(keylist[kid]))
					break;

			if(kid<nkey)
				bitmask |= finger<<(nkey-1 - kid);							// Add to bitmask
			else
				return false;
		}

		return ready=true;
	}


	bool Choices::next(int& key)
	{
		if(not ready or iscan<0)return false;								// Never started or mask completely scanned

		bool found = false;													// Started with iscan = nkey-1

		do
		{
			found = finger & bitmask;

			finger<<=1;	--iscan;											// Prepare for next call
		}
		while(not found and iscan>=0);

		key = iscan+1;

		return found;
	}

	const string make_mask_str(const Keylist& keylist, const bits mask)
	{
		const int& nkey(keylist.nkey);

		string maskstr = string();	bits finger = 1;

		bool started = false;

		if(mask == 0)				   return string(kl_extras[kl_none]);
		if(mask == (finger<<nkey) - 1) return string(kl_extras[kl_all]);

		for(int kid=0; kid<nkey; ++kid)
		{
			if(finger & mask)
			{
				if(started)
					maskstr += string("|");									// Prefix key name by |
				else
					started = true;

				maskstr += string(keylist[nkey-1 - kid]);
			}
			finger<<=1;
		}

		return maskstr;
  	}
}

//----------------------------------------------------------------------------------------------------------------------
// ITEMSPEC
//
namespace Tio
{
	Itemspec dummy_itemspec;

	int make_field_count(const int pclass, const int pcount )
	{
		int fcount = unset_i;

		switch(pclass)
		{
			case pc_scalar:  fcount = 1; 					break;			// Exactly one field
			case pc_range:	 fcount = 2;					break;			// Min, max
			case pc_pair:    fcount = 2; 					break;			// Ordered pair
			case pc_vector:  fcount = 0; 					break;			// Data dependent
			case pc_matrix:  fcount = 2; 					break;			// Matrix rows, cols

			case pc_row:	 if(pcount>0)  fcount = pcount;	break;			// Accept pos pcount
			case pc_keylist: if(pcount>0)  fcount = pcount;	break;			// List for choice or choices

			case pc_choice:  if(pcount>=0) fcount = pcount; break;			// Link choice to keylist
			case pc_choices: if(pcount>=0) fcount = pcount; break;			// Link choices to keylist

			default: fcount = unset_i;						break;
		}
		return fcount;
	}

	Itemspec::Itemspec(const char* pname, const int pclass, const int pcount)
	:name(pname), iclass(pclass), fcount(make_field_count(pclass, pcount))
	{
	}
}

//----------------------------------------------------------------------------------------------------------------------
// ITEMSPECS
//
namespace Tio
{
	Itemspecs dummy_itemspecs;

	bool Itemspecs::isvalid_specs(void)
	{
		if(err) return false;

		for(eitem=0; eitem<(int)nspec; ++eitem)								// Scan all items
        {
	        const Itemspec& ispec(pispecs[eitem]);

	        if(ispec.name==0)  return fail(ep_inoname);						// Valid item fields
	        if(ispec.fcount<0) return fail(ep_badcount);
	        if(ispec.iclass<0 or ispec.iclass>=pc_nclass)
	        	return fail(ep_badclass);

	        const string& iname(ispec.name);								// Unique names

	        for(size_t j=eitem+1; j<nspec; ++j)
	        	if(string(pispecs[j].name) == iname)
	        		return fail(ep_iname);
       }
        return true;
	}

	Itemspecs::Itemspecs(const char* pname, const size_t pnspec, Itemspec* ppispecs)
	:name(pname), nspec(pnspec), pispecs(ppispecs),
	 err(0)
	{
		if(pnspec>0 and ppispecs)
		{
			isvalid_specs();
		}
		else
			err = ep_makespecs;
	}
}

//----------------------------------------------------------------------------------------------------------------------
// CLASS Itemblock
//
namespace Tio
{
	Itemblock dummy_itemblock;

	bool Itemblock::report_fail(void)
	{
		if(err == 0 ) err = ep_unknown;										// Spurious error report

		const char* errmsg = tio_error_message(err);

		const string bname = get_name();
		const char* iname = get_itemname(eitem);

		clog<<"ITEMBLOCK: "<<(bname.empty() ? string("no-name") : bname);
		clog<<tab<<"eitem "; if(iname) clog<<iname; else clog<<eitem;
		clog<<tab<<errmsg;
		clog<<endl;

		return false;
	}

	Itemblock::Itemblock
	(
		Itemspecs& pispecs,													// Mandatory Itemspecs
		const Refkeylists prkeylists										// Optional Keylists
	)
	:keylists(prkeylists.rklists), ispecs(pispecs),
	 err(0), eitem(0)
	{
		if(not ispecs.err)
		{
			items.resize(ispecs.nspec);

			for(size_t i=0; i<ispecs.nspec; ++i)
            {
	            const Itemspec& ispec(ispecs[i]);

				if(ispec.iclass==pc_choice or ispec.iclass==pc_choices)		// Need a keylist for this item
				{
					const size_t iklist = ispec.fcount;						// Index to keylists

					eitem = i;

					if(iklist<keylists.nkeylist)							// Keylists  index in range
					{
						const Keylist& keylist(keylists[iklist]);

						if(not keylist.nkey or not keylist.keys )			// But no keylist
						{ err = ep_noklist; break; }
					}
					else
					{
						err = ep_idkeylist; break;							// Index out of range
					}
				}
            }
		}
		else
		{
			err = ispecs.err;	eitem = ispecs.eitem;						// Inherit Itemspecs error
		}
	}

	Itemspec& Itemblock::find_item(const string& itemname, int& i)
	{
		if(err) return dummy_itemspec;

		for(i=0; i<(int)items.size(); i++ )
		{
			Itemspec& itemspec = ispecs[i];

			if( itemspec.name==itemname)
				return itemspec;
		}

		err = ep_finditem;

		return dummy_itemspec;
	}

	bool Itemblock::make_itemstring(list<string>& itemlist, const int i)	// ONE TABBED string FROM A LIST OF string
	{
		if(err) return false;		eitem = i;

		if(itemlist.empty()) return fail(ep_fieldlist);

		string& istring(items[i]);	istring.clear();						// Parspec::load_item guarantees valid i

		list<string>::iterator it = itemlist.begin();						// First file entry after the item name

		istring += *it++;													// Extract first field

		while(it != itemlist.end())											// Build the tabbed string of fields
			istring += tab + *it++;

		return true;
	}

	const char* Itemblock::get_itemname(const int i)
	{
		eitem = i;

		if(0<=i && i<(int)ispecs.nspec) return ispecs[i].name;

		return not err ? fail_czero(ep_itemindex) : 0;
	}

	const string& Itemblock::get_itemstring(const int pclass, const int i)
	{
		if(err) return dummy_string;	eitem = i;

		if(0<=i && i<(int)ispecs.nspec)
			if(ispecs[i].iclass==pclass)
				return items[i];
			else
				return fail_string(ep_class);
		else
			return fail_string(ep_itemindex);
	}

	bool Itemblock::get_item_keys
	(
		const int pclass, const size_t i,
		string& item, size_t& ilist
	)
	{
		eitem=i;	if(err) return false;

		if(i>=ispecs.nspec) return fail(ep_itemindex);

		item.clear();

		const Itemspec& itemspec(ispecs[i]);								// Retrieve the itemspec
		if(itemspec.iclass != pclass ) return fail(ep_chooser);				// Should be choice/choices

		ilist = itemspec.fcount;											// Retrieve the keylist id from items fcount
		if(ilist>=keylists.nkeylist) return fail(ep_idkeylist);

		item=items[i];

		return true;
	}

	bool Itemblock::get_choice(const int i, int& ichoice)					// Manage a  single choice item
	{
		ichoice = unset_i;	eitem=i;	if(err) return false;

		string choice;	size_t ilist;

		if(not get_item_keys(pc_choice, i, choice, ilist)) return false;	// Retrieve item string & its keylist id

		const Keylist& keylist(keylists[ilist]);
		const int& nkey(keylist.nkey);

		for(ichoice=0; ichoice<nkey; ++ichoice)								// Find a matching choice in the keylist
			if(string(keylist[ichoice])==choice)
				return true;

		return fail(ep_badchoice);
	}

	bool Itemblock::get_choices(const int i, Choices& choices)				// Manage a multiple choice item
	{
		choices.clear();	eitem=i;	if(err) return false;

		string selection;	size_t ilist;

		if(not get_item_keys(pc_choices, i, selection, ilist))
			return false;

		if(selection == string(kl_extras[kl_none]))
		{
			choices.none();	return true;
		}

		if(choices.build_mask(keylists[ilist], selection))
			return true;

		if(selection == string(kl_extras[kl_all]))
		{
			choices.all();	return true;
		}

		return fail(ep_badchoices);
	}
}

//----------------------------------------------------------------------------------------------------------------------
// CLASS Itemblocks
//
namespace Tio
{
	Itemblocks dummy_itemblocks;

	Itemblocks::Itemblocks(const size_t pnblock, Itemblock* pitemblocks)
	:nblock(pnblock), itemblocks(pitemblocks),
	 err(0), eitem(0), eblock(0)
	{
		if(nblock>0 and itemblocks != 0)
			for(eblock=0; eblock<(int)nblock; ++eblock)
            {
	            const Itemblock& iblock(itemblocks[eblock]);

	            const string& ibname(iblock.get_name());

	            if(ibname.empty()) { fail(ep_bzname); break; }				// No empty names in a block sequence

	            err = iblock.get_error(eitem);	if(err) break;				// Block must be error free

	            for(size_t b=eblock+1; b<nblock; ++b)						// Scan remaining blocks
                {
	            	const Itemblock& block(itemblocks[b]);

	            	const string& bname(block.get_name());

	            	if(bname==ibname){ fail(ep_bname); break; }				// Blocks have unique names
                }
            }
		else
			fail(ep_makeblocks);
	}
}

//----------------------------------------------------------------------------------------------------------------------
// PARSPEC ERROR REPORTING
//
namespace Tio
{
	bool Parspec::fail(const int perr)				{ err = perr;	return false; }
	string Parspec::fail_string(const int perr) 	{ err = perr;	return dummy_string; }
	const char* Parspec::fail_czero(const int perr) { err = perr;	return 0; }

	bool Parspec::report_fail(void)
	{
		clog<<"PARSPEC:"<<tio_error_message(err);

		switch(err)
		{
			case ep_missingblock: case ep_blockindex: case ep_blocksize:
				clog<<tab<<"block "<<eblock;
			break;

			default:
			{
				if(eblock<-1 or eblock>=(int)namedblocks.nblock)
				{
					clog<<tab<<"block "<<eblock<<tab<<"item "<<eitem<<tab<<"line "<<eline;

					break;
				}

				const Itemblock& eitemblock = eblock==-1 ? nonameblock : namedblocks[eblock];

				string eblockname(eitemblock.get_name());
				if(eblockname.empty())
					eblockname = eblock==0 ? string("no-name") : string("empty-name");


				string eitemname;

				if(0<=eitem and eitem<(int)eitemblock.size())
				{
					eitemname = string(eitemblock.ispecs[eitem].name);
					if(eitemname.empty())
						eitemname = string("empty-item");
				}

				clog<<tab<<"block["<<eblock<<"] "<<eblockname;
				clog<<tab<<"item["<<eitem<<"] "<<eitemname;
				clog<<tab<<"line "<<eline;
			}
			break;
		}
		clog<<endl;

		return false;
	}

	bool Parspec::report_fail(const int perr)
	{
		err = perr;		return report_fail();
	}
}

//----------------------------------------------------------------------------------------------------------------------
// INTERNAL TOOLS
//
// AN INPUT LINE IS A NON-EMPTY SEQUENCE OF THINGS SEPARATED BY TAB AND TERMINATED BY EOL
//
namespace Tio
{
	bool Parspec::get_next_line(string& line)
	{
		if(err) return false;

		while(getline(pstream, line))
		{
			++eline;													// We have read a new line

			if(line.empty())continue;									// Ignore empty lines
			if(line[0]=='#')continue;									// Ignore comments

			trim(line);

			return pstream.good();
		}

		return fail(ep_getline);
	}

	const char* Parspec::load_blockname(void)
	{
		if(err) return 0;

		string line;

		if(get_next_line(line))											// line excludes end-of-line
		{
			std::istringstream nstream(line, std::ios::in);

			string blockname;  nstream>>blockname;						// First thing is the blockname

			if(line.size()==blockname.size())							// There is only one thing
				return blockname.c_str();
			else
				return fail_czero(ep_syntax);
		}
		else
			return fail_czero(ep_getline);
	}

	bool Parspec::load_item(Itemblock& block)
	{
		using std::istringstream;

		string line;	if(err) return false;

		if(get_next_line(line))
		{
			istringstream lstream(line, std::ios::in);

			int i;  string itemname;  lstream>>itemname;				// First thing is the itemname

			Itemspec& itemspec(block.find_item(itemname, i));			// Identify specification from its name

			err = block.err;	if(err)return false;					// todo what about eitem?

			// BUILD FIELD LIST
			string field;	list<string>fieldlist;						// List of fields detected on this line

			bool fpass = true;											// Compare fields read to specifed field count

			switch(itemspec.iclass)
			{
				case pc_scalar: case pc_range: case pc_pair: case pc_row:
				case pc_keylist:
				{
					while(!lstream.eof())								// Build field list
					{
						lstream>>field;	fieldlist.push_back(field);
					}
					fpass = (int)fieldlist.size()==itemspec.fcount;
				}
				break;

				case pc_choice: case pc_choices: 						// There is just one field
				{														// The specified fieldcount indexes itemspecs
					lstream>>field;	fieldlist.push_back(field);
					fpass = not field.empty();
				}
				break;

				case pc_vector:
				{
					string sdim;	lstream >> sdim;					// Dimensionality

					const int ndim = str_to_int(sdim);

					if(ndim>0)
					{
						itemspec.fcount = ndim+1;
						fieldlist.push_back(sdim);

						while(!lstream.eof())								// Vector components
						{
							lstream>>field;	fieldlist.push_back(field);
						}
						fpass = (int)fieldlist.size()==itemspec.fcount;
					}
					else
						return fail(ep_vector);
				}
				break;

				case pc_matrix:
				{
					string snrow, sncol;	lstream>>snrow>>sncol;			// Matrix order

					const int nrow = str_to_int(snrow), ncol = str_to_int(sncol);

					int col;

					if(nrow>0 && ncol>0)
					{
						itemspec.fcount = nrow*ncol + 2;

						fieldlist.push_back(snrow);
						fieldlist.push_back(sncol);

						for(int r=0; r<nrow; r++)
						if(get_next_line(line))								// Read a row
						{
							istringstream mstream(line); col=0;

							while(!mstream.eof())							// Matrix entries
							{
								mstream>>field;	fieldlist.push_back(field);
								++col;
							}
							if(col!=ncol)break;
						}
						else
							return fail(ep_matrix);
					}
					else
						return fail(ep_matrix);

					fpass = col==ncol;
				}
				break;

				default: return fail(ep_loaditem);
			}

			if(fpass)
				if(block.make_itemstring(fieldlist,i))
					return true;

			return fail(ep_fieldlist);
		}
		return false;
	}

	Itemblock& Parspec::find_namedblock(const char* cname)
	{
		if(cname==0) { err = ep_bzname;  return dummy_itemblock; }

		const string& sname(cname);

		for(size_t r=0; r<namedblocks.nblock; r++)
			if(string(namedblocks.get_blockname(r)) == sname)
				return namedblocks[r];

		err = ep_missingblock;

		return dummy_itemblock;
	}

	bool Parspec::load_block(const char* cname)
	{
		if(err) return false;

		Itemblock& iblock(cname==0 ? nonameblock : find_namedblock(cname));

		if(!err)
		{
			size_t iloaded = 0;

			for(iloaded=0; iloaded<iblock.ispecs.nspec; iloaded++)
			{
				eitem = iloaded;

				if(not load_item(iblock)) break;
			}

			return iloaded == iblock.ispecs.nspec;							// todo why on earth force to read all items
		}
		else
			return false;
	}
}

//----------------------------------------------------------------------------------------------------------------------
// CONSTRUCTION
//
namespace Tio
{
	Parspec::Parspec(Itemblock& pnonameblock, std::ifstream& ppstream)		// A single anonymous row of items
	:err(ep_none), eblock(0), eitem(0), eline(0),
	 nonameblock(pnonameblock), namedblocks(dummy_itemblocks),
	 pstream(ppstream)
	{
		while(true)
		{
			if(!nonameblock.get_name().empty())	{ fail(ep_emptyblockname); 	break; }
			if(nonameblock.size()==0) 			{ fail(ep_blocksize); 		break; }

			err = nonameblock.get_error(eitem);	if(err) break;

			break;
		}
	}

	Parspec::Parspec(Itemblocks& pnamedblocks, std::ifstream& ppstream)		// A sequence of named blocks
	:err(ep_none), eblock(0), eitem(0), eline(0),
	 nonameblock(Itemblock()), namedblocks(pnamedblocks),
	 pstream(ppstream)
	{
		while(true)
		{
			if(namedblocks.empty()){ fail(ep_emptyblocks);     break; }

			err = namedblocks.get_error(eblock,eitem); if(err) break;

			break;
		}
	}
}

//----------------------------------------------------------------------------------------------------------------------
// INTERFACE
//
namespace Tio
{
	bool fail(const size_t err)
	{
		clog<<tio_error_message(err)<<endl;

		return false;
	}

	bool isvalid_file(const char* pfile)
	{
		if(!pfile)return fail(ep_fname);									// No filename

		struct stat pfilestat;												// No file

		if(stat(pfile, &pfilestat)<0)return fail(ep_fexist);

		return true;
	}

	bool load_parspec_header(std::ifstream& fstream, string& name, string& value)
	{
		if(not fstream.good()) return fail(ep_fopen);

		string line; char ctab;

		while(getline(fstream, line))
		{
			if(line.empty())continue;										// Ignore empty lines
			if(line[0]=='#')continue;										// Ignore comments

			std::istringstream lstream(line);

			lstream.unsetf(std::ios::skipws);

			lstream>>name>>ctab>>value;

			if(name.empty() or value.empty() or ctab != '\t') return false;

			return lstream.eof();
		}
		return false;
	}

	bool validate_parspec_header(std::ifstream& fstream, const char* htitle, const char* hvalue)
	{
		string name, value;

		if(not load_parspec_header(fstream, name, value)) return fail(ep_fheader);

		if(name != string(htitle) or value != string(hvalue))
		{
			fstream.close();

			return fail(ep_fheader);
		}

		return true;
	}

	bool Parspec::load(void)
	{
		if(err) return false;

		// LOAD THE STACKS
		while(true)
		{
			eblock = -1;

			if(!nonameblock.ispecs.empty())
				if(!load_block(0))break;

			if(!namedblocks.empty())
				for(size_t s=0; s<namedblocks.nblock; s++)
				{
					eblock = s;	eitem = 0;

					const char* blockname = load_blockname();				// A blockname expected

					if(!load_block(blockname))break;						// Find the specs for the block and load
				}
			break;
		}

		// CLOSE THE FILE AND REPORT ANY LINE ERROR
		bool endfile=pstream.eof();	pstream.close();

		if(endfile)return fail(ep_fend);									// Premature end

		return !err;
	}

	const char* Parspec::get_blockname(void)								// Anonymous block
	{
		if(err) return 0;	eblock = -1;

		return nonameblock.empty() ? fail_czero(ep_anonymousblock) : 0;
	}

	const char* Parspec::get_blockname(const int iblock)					// Named blocks
	{
		if(err) return 0;	eblock = iblock;

		if(nonameblock.empty())
		{
			Itemblock& itemblock(namedblocks[iblock]);

			if(not itemblock.empty()) return itemblock.ispecs.name;
		}

		return fail_czero(ep_namedblocks);
	}

	Itemblock& Parspec::get_itemblock(void)
	{
		if(err) return dummy_itemblock;	eblock = -1;

		return nonameblock.empty() ? dummy_itemblock : nonameblock;
	}

	Itemblock& Parspec::get_itemblock(const int iblock)
	{
		if(err) return dummy_itemblock;

		eblock = iblock;

		return nonameblock.empty() ? namedblocks[iblock] : dummy_itemblock;
	}

	int Parspec::get_error(int& eblockp, int& eitemp, int& elinep)const
	{
		eblockp = eblock;	eitemp = eitem;		elinep = eline;

		return err;
	}

	void Parspec::get_error_names(string& sblock, string& sitem)
	{
		sblock = string("unknown block");	sitem = string("unknown item");

		if(eblock<-1 || eblock>=(int)namedblocks.nblock)return;

		Itemblock& eitemblock = eblock==-1 ? nonameblock : namedblocks[eblock];

		sblock = string(eitemblock.get_name());

		if(0<=eitem && eitem<eitemblock.size())
			sitem  = string(eitemblock.ispecs[eitem].name);
	}
}

//----------------------------------------------------------------------------------------------------------------------
// BLOCK OUTPUT
//
namespace Tio
{
	using Appmaths::Slice;

	enum
	{
		eb_none,
		eb_bstream,
		eb_bindex,eb_beginblock,eb_newblock, eb_colheader, eb_endblock,
		eb_coltext,
		eb_matrix, eb_vector,
		eb_unknown
	};

	const char *left_bracket = " ([<{?", *right_bracket = " )]>}?";

	const string lbra("("), rbra(")"), lban("<"), rban(">"), lbsq("["),rbsq("]"), lbcu("{"), rbcu("}");

	Ovector::Ovector()
	:dmissing(false), vnodata(0), tnodata(string())
	{
		title = tab;	scale = 1;	places = 1;
	}

	Ovector::Ovector(const real pvnodata, const string& ptnodata)
	:dmissing(true), vnodata(pvnodata), tnodata(ptnodata)
	{
		title = tab;	scale = 1;	places = 1;
	}

	Ovector::Ovector(const real pvnodata, const int nplace)
	:dmissing(true), vnodata(pvnodata)
	{
		title = tab;	scale = 1;	places = nplace;

		tnodata = string("?.") + string(places, '?');
	}


	Outblock::Outblock(std::ofstream& pbstream)
	:err(eb_none), blockindex(0), bstream(pbstream)
	{
		if(not bstream.good()) err = eb_bstream;
	}

	Outblock::Outblock(const int pindex, std::ofstream& pbstream)
	:err(eb_none), blockindex(pindex), bstream(pbstream)
	{
		if(not bstream.good()) err = eb_bstream;
	}

	bool Outblock::block_index(void)
	{
		if(err)return false;
			bstream<<hash<<lbsq<<blockindex++<<rbsq<<space;
		return bstream.good() ? true : fail(eb_bindex);
	}

	bool Outblock::start(const int pindex, const string& title)
	{
		if(err) return false;

		blockindex = pindex;

		if(not block_index()) return false;
			bstream<<title<<endl;
		return bstream.good() ? true : fail(eb_beginblock);
	}

	bool Outblock::next(const string& title)
	{
		if(err) return false;

		if(not block_index()) return false;
			bstream<<title<<endl;
		return bstream.good() ? true : fail(eb_newblock);
	}

	bool Outblock::next(const char* title)
	{
		if(err) return false;

		if(not block_index()) return false;
			bstream<<title<<endl;
		return bstream.good() ? true : fail(eb_newblock);
	}


	int Outblock::next_index(void)
	{
		if(err) return unset_i;

		return blockindex++;
	}

	int Outblock::next_index(const int stride)
	{
		if(err) return unset_i;

		if(stride<1) return blockindex;

		const int newindex = blockindex;

		blockindex += stride;

		return newindex;
	}

	bool Outblock::colhead(const char* title, const Slice& crange)
	{
		if(err) return false;

		if(not crange.set) return fail(eb_colheader);

		bstream<<title;
			for(int c=crange.min; c<=crange.max; ++c) bstream<<tab<<c;

		return bstream.good() ? true : fail(eb_coltext);
	}

	bool Outblock::colhead(const string& title, const Slice& crange)
	{
		if(err) return false;

		if(not crange.set) return fail(eb_colheader);

		bstream<<title;
			for(int c=crange.min; c<=crange.max; ++c) bstream<<tab<<c;

		return bstream.good() ? true : fail(eb_colheader);
	}

	bool Outblock::colhead(const string& prefix, int brastyle, const Slice& crange)
	{
		if(err) return false;	if(not crange.set) return fail(eb_colheader);

		if(brastyle<bra_none or brastyle>bra_size)brastyle=bra_size;

		const char lbr = left_bracket[brastyle], rbr = right_bracket[brastyle];

		for(int c=crange.min; c<=crange.max; ++c)
		{
			bstream<<tab<<prefix;

			if(brastyle>0)
				bstream<<lbr<<c<<rbr;
			else
				bstream<<c;
		}
		return bstream.good() ? true : fail(eb_colheader);
	}

	bool Outblock::colhead(int brastyle, int fixstyle, const char* strfix, const Slice& crange)
	{
		if(err) return false;if(not crange.set) return fail(eb_colheader);

		if(brastyle<bra_none or brastyle>bra_size) brastyle = bra_size;
		if(fixstyle<fix_pre  or fixstyle>fix_none) fixstyle = fix_none;

		const char lbr = left_bracket[brastyle], rbr = right_bracket[brastyle];

		for(int c=crange.min; c<=crange.max; ++c)
		{
			bstream<<tab;

			if(fixstyle==fix_pre) bstream<<strfix;

			if(brastyle>bra_none)
				bstream<<lbr<<c<<rbr;
			else
				bstream<<c;

			if(fixstyle==fix_post) bstream<<strfix;
		}
		return bstream.good() ? true : fail(eb_colheader);
	}

	bool Outblock::colhead(int brastyle, int fixstyle, const string& strfix, const Slice& crange)
	{
		return colhead(brastyle, fixstyle, strfix.c_str(), crange);
	}

	bool Outblock::colhead(const char* trow, size_t ncol, const char** tcols)
	{
		if(err or trow==0 or tcols==0) return false;

		bstream<<trow;

		for(size_t c=0; c<ncol; ++c)bstream<<tab<<tcols[c];

		return true;
	}

	bool Outblock::finish(void)
	{
		bstream<<endl<<endl;
		bstream.flush();

		return not err and bstream.good() ? true : failed();
	}

	bool Outblock::failed(void)
	{
		std::clog<<endl;
		std::clog<<"Outblock FAILED"<<endl;
		return false;
	}

	bool Outblock::matrix
	(
		const Omatrix& omat,												// Layout parameters
		const int rshow, const int cshow,									// Whether to show row or column names
		const vector< vector<real> >& mat
	)
	{
		// Matrix row/column indices are transformed to row/column names:
		// name = name.min + index*stride

		if(err) return false;

		if(!omat.rows.set or !omat.cols.set or omat.mplaces<0)				// Layout parameter check
			return fail(eb_matrix);

		if(omat.rstride<=0 or omat.cstride<=0) return fail(eb_matrix);

		const size_t r_size = (omat.rows.max-omat.rows.min)/omat.rstride+1; // Layout/matrix consistency
		const size_t c_size = (omat.cols.max-omat.cols.min)/omat.cstride+1;

		if(r_size==0 or c_size==0) 	return fail(eb_matrix);
		if(mat.empty()) 			return fail(eb_matrix);

		if(omat.transpose)													// Supplied mat is reqd to be transposed
		{
			if(mat.size() != c_size) 	return fail(eb_matrix);
			if(mat[0].size() != r_size)	return fail(eb_matrix);
		}
		else
		{
			if(mat.size() != r_size) 	return fail(eb_matrix);
			if(mat[0].size() != c_size)	return fail(eb_matrix);
		}

		if(cshow)															// Column headings
		{
			if(rshow)bstream<<omat.title;

			for(int col=omat.cols.min; col<=omat.cols.max; col += omat.cstride)
				bstream<<tab<<col;
			bstream<<endl;
		}

		for(size_t r=0; r<r_size; ++r)
        {
			if(rshow==rows_on)												// Display row name
			{
				const int irowname = omat.rows.min + r*omat.rstride;		// Current row name

				if(omat.rplaces==0)											// As is
					bstream<<irowname;

				if(omat.rplaces != 0)										// Scaled
				{
					bstream<<fixed<<setprecision(omat.rplaces);
					bstream<<omat.rzero + irowname*omat.rscale;
				}
			}

			bstream<<fixed<<setprecision(omat.mplaces);						// Display row

			for(size_t c=0; c<c_size; ++c)
			{
				real matrc = omat.transpose ? mat[c][r] : mat[r][c];
				matrc = omat.mzero + omat.mscale*matrc;

				const bool nodata = omat.dmissing and matrc==omat.vnodata;

				if(c==0)
				{
					switch(rshow)
					{
						case rows_off:										// First item
							if(nodata)bstream<<omat.tnodata; else bstream<<matrc;
						break;

						case rows_on:										// Second item following row name
							if(nodata)bstream<<tab<<omat.tnodata; else bstream<<tab<<matrc;
						break;

						default: return false;
					}

					continue;
				}

				if(nodata)bstream<<tab<<omat.tnodata; else bstream<<tab<<matrc;
			}
			bstream<<endl;
        }

		return bstream.good() ? true : fail(eb_matrix);
	}


	bool Outblock::rvector(const Ovector& ovec, const vector<real>& vec)
	{
		if(vec.empty()) return fail(eb_vector);

		const std::streamsize vplaces = bstream.precision(ovec.places);

		bstream<<ovec.title;

		for(size_t i=0; i<vec.size(); ++i)
		{
			const real veci = vec[i];

			if(ovec.dmissing and veci==ovec.vnodata)
			{
				bstream<<tab<<ovec.tnodata;	continue;
			}

			bstream<<tab<<ovec.scale*veci;
		}

		bstream.precision(vplaces);

		return bstream.good() ? true : fail(eb_vector);
	}

	bool Outblock::ivector(const string& title, const vector<int>& vec)
	{
		if(vec.empty()) return fail(eb_vector);

		bstream<<title;

		for(size_t i=0; i<vec.size(); ++i) bstream<<tab<<vec[i];

		return bstream.good() ? true : fail(eb_vector);
	}

	void report_progress(void)
	{
		static int rcount = 0;

		if(rcount++ < 10) cout<<plus;

		if(rcount==10)
		{
			cout<<endl;

			rcount = 0;
		}
	}
}
// END Tio
//----------------------------------------------------------------------------------------------------------------------

/*
 * 			for(size_t c=0; c<c_size; ++c)
			{
				real matrc = omat.transpose ? mat[c][r] : mat[r][c];

				const bool nodata = omat.dmissing and matrc==omat.vnodata;

				if(omat.dmissing and matrc==omat.vnodata)
				{
					bstream<<tab<<omat.tnodata;		continue;
				}

				matrc = omat.mzero + omat.mscale*matrc;

				if(c==0)
				{

					switch(rshow)
					{
						case rows_off:	bstream<<matrc; 		break;
						case rows_on:	bstream<<tab<<matrc;	break;

						default: return false;
					}
				}
				else
					bstream<<tab<<matrc;
			}
 *
 */
