//----------------------------------------------------------------------------------------------------------------------
// Tio	MANAGE TEXT I/O  						Created on 24 Feb 2012						Copyright Geoff Kaniuk 2012
//----------------------------------------------------------------------------------------------------------------------
//  This file is part of Tio.
//
//  Tio is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  Tio is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with Tio.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------------------------------------------------
// A copy of the GNU General Public License is released with this software in the file COPYING
//----------------------------------------------------------------------------------------------------------------------


#include <string>
#include <vector>
#include <list>

#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>

#include "arithmetic.h"
#include "intervals.h"

#ifndef TIO_H_
#define TIO_H_

//----------------------------------------------------------------------------------------------------------------------
// ERROR REPORTING
//
namespace Tio
{
	enum
	{
		ep_none,

		ep_vectordim, ep_matrixdim,											// Errors local to tio.h

		ep_fname, ep_fexist, ep_fopen, ep_fheader, ep_fend, ep_fclose,		// Interface errors

		ep_nerrtioh
	};

	extern const char* tio_error_message(const int e);

	extern std::string dummy_string;
}

//----------------------------------------------------------------------------------------------------------------------
// STRING TOOLS
//
namespace Tio
{
	using std::string;		using std::vector;			using std::list;

	using std::cout;		using std::endl;
	using std::fixed; 		using std::setprecision;

	using Appmaths::real;	using Appmaths::Interval;	using Appmaths::Slice;
	using Appmaths::bits;	using Appmaths::unset_i;

	extern const string tab, hash, hyphen, space;
	extern const string point, colon, hcolon, uscore;
	extern const string equals, plus, star, query;
	extern const string vbar, ampers;

	extern string dummy_string;
	extern const char* dummy_cstring;

	extern int str_to_int(const string& source);
	extern real str_to_real(const string& source);
	extern void trim(string& source);

	extern bool is_whitespace(const char c);
	extern bool parse(const string& source, const char delim, vector<string>& sink);
}

//----------------------------------------------------------------------------------------------------------------------
// KEY LISTS AND CHOICES
//
namespace Tio
{
	template<const size_t csize>											// Array of names
	struct Names
	{
		const char** names;

		Names(const char* pnames[csize]):names(pnames){}

		const char* name(const size_t i)const
		{
			if (csize==0) return "?";

			const size_t ti = i< (csize-1) ? i : csize-1;

			return names[ti];
		}
	};

	struct Keylist															// ARRAY OF KEYS
	{
		const size_t nkey;													// Number of keys;
		const char** keys;

		Keylist():nkey(0), keys(0) {}
		Keylist(const size_t pnkey, const char** pkeys)
		:nkey(pnkey), keys(pkeys)
		{}

		const char* operator[](const int i)const
		{
			if(i<0 or i>=(int)nkey) return dummy_cstring;

			return *(keys+i);
		}

		int find(const char* citem)const
		{
			if(!citem) return unset_i;

			const string item(citem);

			for(size_t k=0; k<nkey; ++k)
				if(item==string(*(keys+k)))
					return k;

			return unset_i;
		}
	};
	extern const Keylist dummy_keylist;

	struct Keylists															// ARRAY OF KEYLISTS
	{
		const size_t nkeylist;												// Number of Keylist members
		const Keylist* keylists;

		Keylists():nkeylist(0), keylists(0){}
		Keylists(const size_t pnklist, const Keylist* pkeylists)
		:nkeylist(pnklist), keylists(pkeylists)
		{}

		const Keylist& operator[](const int i)const
		{
			if(i<0 or i>=(int)nkeylist) return dummy_keylist;

			return *(keylists+i);
		}
	};
	extern const Keylists dummy_keylists;

	struct Refkeylists														// REFERENCE TO KEYLISTS
	{
		const Keylists& rklists;

		Refkeylists():rklists(dummy_keylists){}
		Refkeylists(const Keylists& pklists):rklists(pklists){}
	};
	extern const Refkeylists dummy_refkeylists;

	class Choices
	{
		private:
			size_t nkey;													// Number of keys
			bits bitmask;													// The current mask from build_mask
			bits finger;													// A bit position in bitmask

			int iscan;														// Manage scanning the bitmask
			bool ready;														// Ready for use

		public:
			Choices(void)
			:bitmask(0), finger(0), iscan(-1), ready(false)
			{}

			bool build_mask( const Keylist& keylist,const string& kselects);

			bool get_ready(void)const  { return ready; }
			bits get_bitmask(void)const { return bitmask; }

			void clear(void){ nkey=iscan=0; finger=bitmask=0; 	  ready = false; }
			void all(void)  { bitmask = (bits)((int)(1<<nkey)-1); ready = true; }
			void none(void) { iscan=-1; 	finger=bitmask=0; 	  ready = false; }
			void start()    { iscan=nkey-1; finger=1; }

			bool next(int& key);
	};

	extern const string make_mask_str(const Keylist& keylist, const bits mask);
}

//----------------------------------------------------------------------------------------------------------------------
// SPECIFICATION OF THE INPUT FORMAT
//
namespace Tio
{
	using Appmaths::unset_i;

	// TEXT FILE INPUT: AN INPUT LINE IS A NON-EMPTY SEQUENCE OF THINGS SEPARATED BY TAB AND TERMINATED BY EOL
	// THERE ARE 9 POSSIBLE TYPES OF INPUT ITEMS - TAB IS REPRESENTED HERE BY ~
	// THE FORMAT IS NAME~THING~THING
	// # COMMENT					# The line is ignored
	// STACKNAME					# The line is a header for the sequence of items following the line
	// SCALAR~item					# exactly one item in the text string
	// PAIR~one~two					# exactly two items of same type
	// ARRAY~a1~a2~a3				# program knows how many - all of same type
	// VECTOR~n~v1~v2~  ... vn		# first item after name is number of items. Then exactly those items of same type.
	// MATRIX~rows~cols				# defines matrix order for following sequence of data vectors
	// d00~d01~d02
	// d10~d11~d12
	// :KEYS~keya~keyb				# Defines an option list
	// CHOICE~keyx					# The required choice from a list
	// &CHOICES~keya~keyb			# Defines a multiple choice option list "none" and "all" automatically added
	// CHOICES~keycombination		# Choices separated by |

	enum																	// INPUT TYPES
	{
		pc_empty,
		pc_scalar, pc_range,  pc_pair, pc_row, pc_vector, pc_matrix,
		pc_keylist,pc_choice, pc_choices,
		pc_nclass
	};

	struct Itemspec															// BASIC UNIT OF INPUT
	{
		const char* name;
		int iclass;															// The class pc_scalar ...
		int fcount;															// Expected number of data fields

		Itemspec():name(0), iclass(unset_i), fcount(unset_i) {};
		Itemspec(const char* pname, const int pclass, const int pcount=0);
	};
	extern Itemspec dummy_itemspec;

	struct Pmodel : public Itemspec											// PARAMETER MODEL
	{
		Pmodel(const char* pname):Itemspec(pname, pc_row, 3){}				// "name"~setpoint~rate~residual
	};

	struct Iscalar : public Itemspec
	{
		Iscalar(const char* pname):Itemspec(pname, pc_scalar){}
	};

	struct Ipair : public Itemspec
	{
		Ipair(const char* pname):Itemspec(pname, pc_pair){}
	};

	struct Irange : public Itemspec
	{
		Irange(const char* pname):Itemspec(pname, pc_range){}
	};

	struct Ichoice : public Itemspec
	{
		Ichoice(const char* pname, const int idkeys):Itemspec(pname, pc_choice, idkeys){}
	};

	struct Ichoices : public Itemspec
	{
		Ichoices(const char* pname, const int idkeys):Itemspec(pname, pc_choices, idkeys){}
	};


	struct Itemspecs														// A NAMED SEQUENCE OF INPUT ITEMS
	{
		const char* name;
		const size_t nspec;
		Itemspec* pispecs;

		int err, eitem;

		Itemspec& operator[] (const size_t i)
		{ return i<nspec ? pispecs[i] : dummy_itemspec; }

		bool fail(const int perr) { err=perr; return false; }

		bool isvalid_specs(void);
		bool empty(void) { return nspec==0 or pispecs==0; }
		Itemspecs():name(0), nspec(0),pispecs(0), err(0){}
		Itemspecs(const char* pname, const size_t pnspec, Itemspec* ppispecs);
	};
	extern Itemspecs dummy_itemspecs;

	class Parspec;

	class Itemblock															// NAMED COLLECTION OF ITEMS
	{
		friend class Parspec;

		private:
			const Keylists& keylists;										// Keys for pc_choice or pc_choices items

			Itemspecs& ispecs;												// Non-empty specification
			vector<string>items;											// Each items[i] is a string with tab delims

			int err, eitem;													// Error reporting

		private:
			int size(void)const { return (int)ispecs.nspec; }

			Itemspec& find_item(const string& itemname, int& i);

			bool make_itemstring(list<string>& itemlist, const int i);

			bool		  fail(const int perr)		  { err = perr;	return false; }
			const char*   fail_czero(const int perr)  { err = perr; return 0; }
			const string& fail_string(const int perr) { err = perr;	return dummy_string; }

			const string& get_itemstring(const int pclass, const int i);

			bool get_item_keys
			(
				const int pclass, const size_t i,
				string& item, size_t & ilist
			);

			size_t get_nkeylist(void)const
			{ return keylists.nkeylist; }

			size_t get_nkey(const size_t l)const
			{ return l<keylists.nkeylist ? keylists[l].nkey : 0; }

			const char* get_key(const size_t l, const size_t k)const
			{
				if(l>=keylists.nkeylist) return 0;

				const Keylist& lkeys(keylists[l]);

				return k<lkeys.nkey ? lkeys[k] : 0;
			}

		public:
			Itemblock()
			:keylists(dummy_keylists), ispecs(dummy_itemspecs),
			 err(0), eitem(0)
			{}

			Itemblock(Itemspecs& specs, const Refkeylists prkeylists = dummy_refkeylists);

			string get_name(void)const
			{
				return ispecs.name ? string(ispecs.name) : string();
			}

			int size(void) { return ispecs.nspec; }
			int get_error(void)const { return err; }
			int get_error(int& eitemp)const { eitemp = eitem;	return err; }
			bool report_fail(void);
			bool empty(void)const { return items.empty() or ispecs.nspec==0; }
			bool get_choice(const int i, int& choice);
			bool get_choices(const int i, Choices& choices);

			const char* get_itemname(const int i);

			bool set_itemstring(const size_t i, const char* itemstr)
			{
				if(i>=items.size()) return false;

				items[i] = string(itemstr);

				return true;
			}

			const char* get_itemstring(const size_t i)
			{
				if(i>=items.size()) return 0;

				return items[i].c_str();
			}

			template<class D>
			bool get(const int it, D& sink)
			{
				eitem = it;

				std::istringstream streami(get_itemstring(pc_scalar, it));

				if(!err){ streami>>sink; return true; }

				return false;
			}

			template<class D>
			bool get(const int it, D& sinka, D& sinkb)
			{
				eitem = it;

				std::istringstream streami(get_itemstring(pc_pair, it));

				if(!err){ streami>>sinka>>sinkb; return true; }

				return false;
			}

			template<class D>
			bool get(const int it, Interval<D>& sink)
			{
				eitem = it;

				std::istringstream streami(get_itemstring(pc_range, it));

				D smin, smax;

				if(!err)streami>>smin>>smax; else return false;

				sink = Interval<D>(smin, smax);								// Verifies smin<=smax

				if(sink.set) return true;

				return false;
			}

			template<class D>
			bool get(const int it, D* sink)									// We don't know size of array sink
			{
				eitem = it;

				std::istringstream streami(get_itemstring(pc_row, it));

				if(!err)
				{
					const int nfield = ispecs[it].fcount;					// This is the specified number

					for(int i=0; i<nfield; i++)
						streami>>sink[i];									// Trust caller meets the spec!

					return true;
				}

				return false;
			}

			template<class D>
			bool get_vector(const int it,  vector<D>& sink)
			{
				eitem = it;

				std::istringstream streami(get_itemstring(pc_vector, it));

				if(!err)
				{
					int ndim;	streami>>ndim; if(ndim<=0) return fail(ep_vectordim);

					sink.resize(ndim);

					for(int i=0; i<ndim; i++)streami>>sink[i];

					return true;
				}

				return false;
			}

			template<class D>
			bool get_matrix(const int it,  vector< vector<D> >& sink)
			{
				eitem = it;

				std::istringstream streami(get_itemstring(pc_matrix, it));

				if(!err)
				{
					int nrow, ncol;	streami>>nrow>>ncol;

					if(nrow<=0 or ncol<=0) return fail(ep_matrixdim);

					sink.resize(nrow);

					for(int i=0; i<nrow; i++)
					{
						vector<D>& mati = sink[i];	mati.resize(ncol);

						for(int j=0; j<ncol; j++)streami>>mati[j];
					}
					return true;
				}
				return false;
			}
	};
	extern Itemblock dummy_itemblock;

	struct Itemblocks														// ARRAY OF ITEMBLOCK
	{
		size_t nblock;														// Number of blocks
		Itemblock* itemblocks;

		int err, eitem, eblock;												// Errors might be inherited

		bool fail(const int perr) { err=perr; return false; }

		int get_error(int& eblockp, int& eitemp)
		{ eblockp = eblock;	eitemp = eitem; return err; }

		bool empty(void) { return nblock==0 or itemblocks==0; }

		Itemblocks():nblock(0), itemblocks(0){}
		Itemblocks(const size_t pnblock, Itemblock* pitemblocks);

		Itemblock& operator[](const size_t b)
		{
			eblock = b;

			return b<nblock ? *(itemblocks+b) : dummy_itemblock;
		}

		string get_blockname(const size_t b)const
		{ return b<nblock ? itemblocks[b].get_name() : string(); }
	};
	extern Itemblocks dummy_itemblocks;
}


//----------------------------------------------------------------------------------------------------------------------
// CLASS PARSPEC - READ THE PARAMETER SPECIFICATIONS FILE
//
namespace Tio
{
	class Parspec															// DATA FILE IS A SEQUENCE OF ITEMSPECS
	{
		public:
			int err;

		private:
			int eblock, eitem, eline;										// Error reporting

			Itemblock   nonameblock;										// Anonymous block
			Itemblocks& namedblocks;										// Sequence of named blocks

			std::ifstream& pstream;											// Parameter file stream

		private:
			bool fail(const int perr);
			string fail_string(const int perr);
			const char* fail_czero(const int perr);

			Itemblock& find_namedblock(const char* cname);

			const char* load_blockname(void);
			bool load_block(const char* blockname);
			bool load_item(Itemblock& block);

			bool get_next_line(string& line);

		public:
			bool report_fail(const int perr);
			bool report_fail(void);

			bool isvalid_file(void);

			Parspec(Itemblock& pnonameblock, std::ifstream& ppstream);		// A single anonymous row of items
			Parspec(Itemblocks& pitemblocks, std::ifstream& ppstream);		// A sequence of named blocks

			size_t namedblock_size(const size_t iblock)const
			{
				if(iblock<namedblocks.nblock)
					return namedblocks[iblock].ispecs.nspec;

				return 0;
			}

			bool load(void);

			int get_error(int& eblockp, int& eitemp, int& elinep)const;

			void get_error_names(string& sblock, string& sitem);

			const char* get_blockname(void);
			const char* get_blockname(const int iblock);

			Itemblock& get_itemblock(void);
			Itemblock& get_itemblock(const int iblock);

			const size_t size_namedblocks(void)
			{ return namedblocks.nblock; }
	};

	extern bool fail(const size_t err);
	extern bool isvalid_file(const char* pfile);
	extern bool load_parspec_header(std::ifstream& fstream, string& name, string& value);
	extern bool validate_parspec_header(std::ifstream& fstream, const char* htitle, const char* hvalue);
}

//----------------------------------------------------------------------------------------------------------------------
// BLOCK OUTPUT
//
namespace Tio
{
	enum 																	// Brackets
	{
		bra_none,
		bra_round, bra_square, bra_angle, bra_curl,
		bra_size
	};

	enum																	// Decorate
	{
		fix_pre, fix_post, fix_none
	};

	enum{ rows_off=0, cols_off=0, rows_on=1, cols_on=1 }; 					// Display row or column names

	class Outblock;

	class Omatrix															// Parameters for matrix display
	{
		friend class Outblock;

		private:
			bool dmissing;													// Whether representing missing data
			string tnodata;													// Text for missing data
			real vnodata;													// Value for missing data

		public:
			string title;													// Header for rows/cols
			Slice  rows, cols;												// Row and col names are numeric
			int rstride, cstride;											// Increments in row,column names

			real rscale, rzero;												// Display rzero + rowname*rscale
			int  rplaces;													// Row name precision


			real mscale;													// Scale matrix entries
			real mzero;														// Display mzero + mscale*mij
			int  mplaces;													// Matrix entry precision

			bool transpose;													// Whether supplied mat is to be transposed

		public:
			Omatrix()
			:dmissing(false),
			 rstride(1), cstride(1),
			 rscale(0), rzero(0), rplaces(0),
			 mscale(1), mzero(0), mplaces(1),
			 transpose(false)
			{}

			void set_missing_data(const real pvnodata, const string& ptnodata)
			{
				vnodata = pvnodata;	tnodata = ptnodata;		dmissing = true;
			}

			void set_missing_data(const int pmplaces, const real pvnodata)
			{
				mplaces = pmplaces;		vnodata = pvnodata;		if(mplaces<1) return;
				tnodata = string(mplaces+2, '?'); tnodata[1] = '.';

				dmissing = true;
			}

			void reset_alldata(void) { dmissing=false; }
	};

	class Ovector															// Embed a vector in line output
	{
		friend class Outblock;

		private:
			bool dmissing;													// Whether representing missing data
			real vnodata;													// Value for missing data
			string tnodata;													// Text for missing data

		public:
			string title;													// Precedes first element
			real scale;														// Scale matrix entries
			int places;														// Fixed point precision

		public:
			Ovector();
			Ovector(const real pvnodata, const string& ptnodata);
			Ovector(const real pvnodata, const int nplace);
	};


	class Outblock
	{
		private:
			int err;														// Report output stream error

			int blockindex;													// Sequential indexing

		private:
			bool fail(const int per){ err = per;  return false; }
			bool block_index(void);

		public:
			std::ofstream& bstream;

			void close(void) { bstream.close(); }

		public:
			Outblock(std::ofstream& pbstream);
			Outblock(const int pindex, std::ofstream& pbstream);

			bool start(const int pindex, const string& title);
			bool next(const char* title);
			bool next(const string& title);
			int  next_index(const int stride);
			int  next_index(void);

			int  get_blockindex(void)const { return blockindex; }

			bool colhead(const char* title, const Slice& crange);
			bool colhead(const string& title, const Slice& crange);
			bool colhead(const string& prefix, int brastyle, const Slice& crange);
			bool colhead(int brastyle, int fixstyle, const char* strfix, const Slice& crange);
			bool colhead(int brastyle, int fixstyle, const string& strfix, const Slice& crange);
			bool colhead(const char* trow, size_t ncol, const char** tcols);

			bool finish(void);
			bool failed(void);

			bool matrix
			(
				const Omatrix& omat,
				const int rshow, const int cshow,
				const vector< vector<real> >& mat
			);

			bool rvector(const Ovector& ovector, const vector<real>& vec);

			bool ivector(const string& title,const vector<int>& vec);
	};

	extern void report_progress(void);
}
#endif
// END Tio
//----------------------------------------------------------------------------------------------------------------------
